# @Time : 2021/5/24 15:46
# @Author : ShaoBaoBaoEr
# @Site :
# @File : neo4j_engine.py
# @Software: PyCharm
import Levenshtein
import py2neo
import networkx as nx
import pandas as pd
import os
from .const import *

from .exceptions import Neo4jEmptyError, Neo4jInitFormatError, Neo4jNodeListIndexError, Neo4jQuickCodeGenerationError
import logging

logger = logging.getLogger(__name__)

from typing import Dict, List, Union, Set, Tuple
from copy import deepcopy
from argparse import Namespace
from .helper import StringMatcher
from queue import Queue
from .wrapper import SortWrapper, DeprecatedWrapper, UnionWrapper
from urllib.parse import urlparse
import re
from collections import OrderedDict
from .neo4j_defauilt_config import NEO4J_DEFAULT_CONFIG


#
# class Neo4jGremlinEngine(object):
#     """"
#     The base engine of gremlin for neo4j
#
#     However the gremlin failed to support neo4j 4.x
#
#
#     """
#
#     def __init__(self, _graph: py2neo.Graph = None):
#         self._graph = _graph


class Neo4jCyberEngine(object):
    """

    各种函数的定义方法

    exists_*/has_* -> 返回布尔值，所需参数永远为node
    get_* -> 返回的是node，所需的参数永远为node或node_id
    find_* -> 返回的是 List[Node]，所需的参数永远为node
    search_* -> 返回的是 List[Node] or Node， 所需的参数

    copy_* -> copy node from current graph and do operations

    """

    def __init__(self, _graph: py2neo.Graph or Dict or Namespace = NEO4J_DEFAULT_CONFIG, syn_flag: bool = None):
        """

        :param _graph: 图实例初始化参数
        """
        self.py2neo_version = py2neo.__version__
        self.graph_map = {}
        if isinstance(_graph, Dict) or isinstance(_graph, Namespace):
            if isinstance(_graph, Dict):
                self.graph_map = _graph
                # {
                #     "NEO4J_HOST": _graph.get("NEO4J_HOST"),
                #     "NEO4J_USERNAME": _graph.get("NEO4J_USERNAME"),
                #     "NEO4J_PASSWORD": _graph.get("NEO4J_PASSWORD"),
                #     "NEO4J_PORT": _graph.get("NEO4J_PORT"),
                #     "NEO4J_PROTOCOL": _graph.get("NEO4J_PROTOCOL"),
                #     "NEO4J_DATABASE": _graph.get("NEO4J_DATABASE"),
                # }
            else:
                self.graph_map = {
                    "NEO4J_HOST": _graph.host,
                    "NEO4J_USERNAME": _graph.username,
                    "NEO4J_PASSWORD": _graph.password,
                    "NEO4J_PORT": _graph.port,
                    "NEO4J_PROTOCOL": _graph.protocol,
                    "NEO4J_DATABASE": _graph.database,
                }
            self.neo4j_graph = Neo4jCyberEngine.set_graph(self.graph_map)
        elif isinstance(_graph, py2neo.Graph):
            self.neo4j_graph = _graph
        else:
            raise Neo4jInitFormatError(_graph)

        if isinstance(_graph, py2neo.Graph):
            __graph_instance = urlparse(re.findall(r'Graph\(\'(.*?)\'\)', self.neo4j_graph.__str__())[0])
            # raise NotImplementedError("_graph object has not implement yet")
            self.graph_map = {
                "NEO4J_HOST": __graph_instance.hostname,
                "NEO4J_USERNAME": __graph_instance.username,
                "NEO4J_PASSWORD": __graph_instance.password,  # CHANGE HERE
                "NEO4J_PORT": __graph_instance.port,
                "NEO4J_PROTOCOL": __graph_instance.scheme,
                "NEO4J_DATABASE": 'neo4j',  # CHANGE HERE
            }
        assert isinstance(self.graph_map, Dict) or isinstance(self.graph_map, Namespace)
        assert self.graph_map is not None, "[*] back up for graph map , if met NotImplement Error ; Restart Map for recomputing"
        self.syn_flag = syn_flag

        self.__cache = {}  # type:Dict[Dict[int,str]]
        self.__node_itself_cache = {}

        self.__register_lambda_functions()
        self.__class_method = {i for i in self.__dir__() if not i.__str__().startswith("_")}

    def __normalize(self, x):
        if x is None:
            return None
        return x.strip('\'').strip('\"')

    def __register_lambda_functions(self):
        """
        TYPE_STRING = 'string'
        TYPE_DOUBLE = 'double'
        TYPE_INTEGER = 'integer'
        TYPE_BOOL = 'bool'
        """
        self.get_string_code = lambda x: self.__normalize(x[NODE_CODE]) if self.__node_type_assertion(x,
                                                                                                      TYPE_STRING) else None
        self.get_integer_code = lambda x: x[NODE_CODE].__str__() if self.__node_type_assertion(x,
                                                                                               TYPE_INTEGER) else None
        self.get_double_code = lambda x: x[NODE_CODE].__str__() if self.__node_type_assertion(x, TYPE_DOUBLE) else None
        self.get_bool_code = lambda x: x[NODE_CODE].__str__() if self.__node_type_assertion(x, TYPE_BOOL) else None

    @staticmethod
    def set_graph(feeder: dict or Namespace):
        """

        :param dict or Namespcae feeder: load from config
        :return:
        """
        try:
            if isinstance(feeder, dict):
                NEO4J_HOST = feeder.get("NEO4J_HOST")
                NEO4J_USERNAME = feeder.get("NEO4J_USERNAME")
                NEO4J_PASSWORD = feeder.get("NEO4J_PASSWORD")
                NEO4J_PORT = feeder.get("NEO4J_PORT")
                NEO4J_PROTOCOL = feeder.get("NEO4J_PROTOCOL")
                NEO4J_DATABASE = feeder.get("NEO4J_DATABASE")
            elif isinstance(feeder, Namespace):
                NEO4J_HOST = feeder.host
                NEO4J_USERNAME = feeder.username
                NEO4J_PASSWORD = feeder.password
                NEO4J_PORT = feeder.port
                NEO4J_PROTOCOL = feeder.protocol
                NEO4J_DATABASE = feeder.database
            else:
                raise KeyError()
        except KeyError as e:
            raise Neo4jInitFormatError("missing key !" + e.__str__())
        g = py2neo.Graph(f"{NEO4J_PROTOCOL}://{NEO4J_HOST}:{NEO4J_PORT}", user=NEO4J_USERNAME,
                         password=NEO4J_PASSWORD)
        return g

    def search_deep_parent_node(self, _node: py2neo.Node, depth: int = 1, except_type: List[str] = None) \
            -> py2neo.Node:
        """
        搜索节点的特定 parent node，可设定深度与类型

        :param _node:
        :param depth:
        :param except_type:
        :return:
        """
        if except_type is None:
            except_type = []
        assert depth >= 1
        _node_cp = None
        for i in range(depth):
            _node_cp = self.find_parent_nodes(_node)
            if _node_cp.__len__() == 0:
                return _node
            if _node_cp.__len__() == 1 and _node_cp[0][NODE_TYPE] in except_type:
                return _node_cp[0]
            else:
                _node = _node_cp[0]
        return _node

    def search_deep_child_node(self, _node: py2neo.Node, depth: int = 1, include_type: List[str] = None) \
            -> py2neo.Node:
        """
        搜索节点的特定 child node，可设定深度与类型

        :param include_type:
        :param _node:
        :param depth:
        :return:
        """
        if include_type is None:
            include_type = []
        assert depth >= 1
        _node_cp = None
        for i in range(depth):
            _node_cp = self.find_child_nodes(_node)
            if _node_cp.__len__() == 0:
                return _node
            if _node_cp.__len__() == 1 and _node_cp[0][NODE_TYPE] in include_type:
                return _node_cp[0]
            else:
                _node = _node_cp[0]
        return _node

    def search_child_nodes(self, _node: py2neo.Node, include_type: list = None, INVERT_TYPE_FLAG=False) \
            -> List[py2neo.Node]:
        """
        搜索一个节点的所有子节点，查找（不是）特定类型的节点

        :param _node:
        :param include_type: 锁定为特定类型
        :param INVERT_TYPE_FLAG: 将搜索原语从是改为不是
        :return:
        """
        child_nodes = self.find_child_nodes(_node)
        result = []
        for child_node in child_nodes:
            if include_type is None:
                result.append(include_type)
            else:
                if not INVERT_TYPE_FLAG:
                    if child_node[NODE_TYPE] in include_type:
                        result.append(child_node)
                else:
                    if child_node[NODE_TYPE] not in include_type:
                        result.append(child_node)
        return result

    def find_ast_root_node_in_line(self, lineno: int, **kwargs) -> List[py2neo.Node]:
        """
        specified the ast root node in one line,
        """
        fileid = kwargs.pop(NODE_FILEID, None)
        funcid = kwargs.pop(NODE_FUNCID, None)
        search_term = {
            NODE_LINENO: lineno,
            NODE_CHILDNUM: 0,  # ROOT节点必然是根节点？
        }

        if fileid is None and funcid is None:
            logger.warning("Please give some hint , otherwise ,the result will be extremely time consuming!")
        if fileid is not None:
            search_term[NODE_FILEID] = fileid
        if funcid is not None:
            search_term[NODE_FUNCID] = funcid

        result = set()
        for _node in self.match(LABEL_AST, **search_term):
            result.add(self.get_ast_root_node(_node))
        result = list(i for i in result if i is not None)
        return sorted(result, key=lambda x: x.identity)

    def get_ast_root_node(self, node: py2neo.Node) -> py2neo.Node:
        """
        specify search method depend on the first parent that contain CFG relation ship

        :param node: 给定一个node，返回其AST的根节点（不一定在同一行）
        :return : None or Node
        """
        # 逆向DFS
        assert node is not None, logger.warning('[-] Input node must not be none ,recheck your code logic')
        # special handler
        # TODO FIX BUGS
        if node[NODE_TYPE] in {TYPE_FUNC_DECL, TYPE_PARAM_LIST, }:
            return node
        # 对某些节点的特殊处理如下：
        # AST_IF，ROOT节点为其条件语句
        if node[NODE_TYPE] in {TYPE_IF}:
            return self.get_child_node(self.get_child_node(node))
        elif node[NODE_TYPE] in {TYPE_IF_ELEM}:
            node = self.get_child_node(node, ignore_error_flag=True)
            if node is None:
                raise NotImplementedError()
            else:
                return node
        # AST_WHILTE,ROOT节点为其条件语句
        elif node[NODE_TYPE] in {TYPE_WHILE}:
            return self.get_child_node(node)
        elif node[NODE_TYPE] in {TYPE_SWITCH_CASE}:
            return self.get_child_node(self.get_parent_node(self.get_parent_node(node)))
        elif self.get_parent_node(node) in {TYPE_SWITCH_CASE}:
            return self.get_child_node(self.get_parent_node(self.get_parent_node(self.get_parent_node(node))))

        while not self.has_cfg(node) and node is not None:
            _node = self.get_parent_node(node, ignore_error_flag=True)
            if _node is None:
                logger.fatal(f"not reachable ; check alg or debug parent node of {node}")
                return
            node = _node
        return node

    def __get_line_root_node_with_cfg(self, node):
        if node is None or node['lineno'] is None:
            return None
        lineno = node['lineno']
        while True:
            if self.has_cfg(node):
                return node
            else:
                parent_node = self.get_parent_node(node, ignore_error_flag=True)
            if parent_node[NODE_LINENO] is None:
                break
            elif parent_node[NODE_LINENO] != lineno:
                break
            else:
                node = parent_node
        return node

    def get_line_root_node(self, node):
        return self.__get_line_root_node_old(node)

    def __get_line_root_node_old(self, node):
        """
        find the line root node
        the current function is very week

        :param node:
        :return:
        """
        if node is None or node['lineno'] is None:
            return None
        lineno = node['lineno']
        while True:
            parent_node = self.get_parent_node(node, ignore_error_flag=True)
            if parent_node['lineno'] is None:
                break
            if parent_node['lineno'] != lineno:
                break
            else:
                node = parent_node
        # DEFINE AST_WHILE AST_IF 's line root node as the conditional node
        # if node[NODE_TYPE] == TYPE_IF or node[NODE_TYPE] == TYPE_WHILE or node[NODE_TYPE] == TYPE_DO_WHILE:
        #     node = \
        #         self.get_ith_child_node(node, 0)
        return node

    def clear_cache(self):
        """
        多图切换的时候，方便清空内存

        :return:
        """
        self.__node_itself_cache = {}
        return True

    def match(self, *args, **kwargs):
        """
        neo4j graph match的接口

        :param kwargs:
        :return:
        """
        return self.neo4j_graph.nodes.match(*args, **kwargs)

    def match_relationship(self, *args, **kwargs):
        """
        neo4j relationship match的接口

        :param args:
        :param kwargs:
        :return:
        """
        return self.neo4j_graph.relationships.match(*args, **kwargs)

    def match_first(self, *args, **kwargs):
        """
        match first node

        :param args:
        :param kwargs:
        :return:
        """
        return self.match(*args, **kwargs).first()

    def __check_plugin_active(self, plugin_name):
        return [i for i, in self.query(f"return {plugin_name}.version()")]

    def check_apoc_active(self):
        return bool(self.__check_plugin_active("apoc").__len__())

    def check_is_global_dim(self, n):
        match = "MATCH (:AST{id:%d})-->(:AST{type:'AST_VAR'})-->(x:AST{type:'string'}) return x" % n['id']
        test = self.neo4j_graph.run(match)
        for t in test:
            if t['x']['code'] in ['_GET', '_POST', '_SERVER', '_REQUEST', '_COOKIES', '_FILES']:
                return True
        return False

    def create(self, _subgraph):
        """
        neo4j graph create interface ; call the create interface ; and auto commit it

        :param _subgraph:
        :return:
        """
        self.neo4j_graph.create(_subgraph)
        return True

    def create_node_if_not_exists(self, label, **kwargs) -> py2neo.Node:
        """
        create node if i not exists and return search node

        :param label   node label
        :param kwargs  node property
        :return: node
        """
        if not self.exists(label, **kwargs):
            self.create(py2neo.Node(label, **kwargs))
        return self.match_first(label, **kwargs)

    def create_relationship_if_not_exists(self, start_node, end_node, label, **kwargs):
        """
        create relationship if it not exists and return search relationship

        :param start_node:
        :param end_node:
        :param label:
        :param kwargs:
        :return:
        """
        if self.neo4j_graph.relationships.match([start_node, end_node], label, **kwargs).first() is None:
            self.create(
                py2neo.Relationship(start_node, label, end_node, **kwargs)
            )
        return self.neo4j_graph.relationships.match([start_node, end_node], label, **kwargs).first()

    def query(self, *args, read_only=True, verbose=True, **kwargs):
        # if verbose:
        #     logger.info(args)
        # store result ?
        return self.neo4j_graph.run(*args, **kwargs)

    def exists(self, *args, **kwargs):
        return self.neo4j_graph.nodes.match(*args, **kwargs).first() is not None

    def delete(self, subgraph):
        self.neo4j_graph.delete(subgraph)

    def to_digraph(self, _subgraph: py2neo.Subgraph):
        """
        convert query result to digraph
        TODO THIS FUNCTION IS NOT COMPLETE AT ALL

        :param _subgraph:
        :return:
        """
        df = pd.DataFrame(_subgraph.data())

        G = nx.DiGraph()
        G.add_nodes_from(list(set(list(df['child']) + list(df.loc['parent']))))

        # Add edges

        tuples = [tuple(x) for x in df.values]
        G.add_edges_from(tuples)
        G.number_of_edges()

        # Perform Graph Drawing
        # A star network  (sort of)
        nx.draw_networkx(G)
        # plt.show()

    def get_graph(self) -> py2neo.Graph:
        """
        返回neo4j graph 实例

        :return:
        """
        return self.neo4j_graph

    def get_parent_node(self, _node: py2neo.Node, ignore_error_flag=False) -> Union[py2neo.Node, None]:
        """
        get first parent node

        :param ignore_error_flag:
        :param _node:
        :return:
        """
        return self.get_ith_parent_node(_node, ignore_error_flag=ignore_error_flag)

    def get_specify_parent_node(self, _node: py2neo.Node,
                                specify_type: set = FUNCTION_CALL_TYPES | FUNCTION_DECLARE_TYPES) -> Union[
        py2neo.Node, None]:
        """
        get the specified parent node with specified type

        :param specify_type:
        :param _node:
        :return:
        """
        if _node[NODE_TYPE] in specify_type:
            return _node
        __node = self.get_parent_node(_node)
        if __node[NODE_TYPE] in specify_type:
            return __node
        elif __node[NODE_TYPE] in {TYPE_STMT_LIST}:
            logger.warning("get specify node error ,get EXIT specifier")
            return None
        else:
            return self.get_specify_parent_node(__node, specify_type=specify_type)

    def get_parent_block(self, _node: py2neo.Node) -> py2neo.Node:
        """
        get the parent block(STMT) node

        :param _node:
        :param ignore_error_flag:
        :return:
        """
        if _node[NODE_TYPE] == TYPE_STMT_LIST: return _node
        parent_node = self.get_parent_node(_node)
        if parent_node[NODE_TYPE] != TYPE_STMT_LIST:
            return self.get_parent_block(_node)
        else:
            return parent_node

    def get_top_block(self, _node: py2neo.Node) -> py2neo.Node:
        """
        get the parent block(STMT) node

        :param _node:
        :param ignore_error_flag:
        :return:
        """
        parent_node = self.get_parent_node(_node)
        if self.get_parent_node(parent_node)[NODE_TYPE] not in {TYPE_FUNC_DECL, TYPE_CLASS, TYPE_TOPLEVEL}:
            return self.get_top_block(self.get_parent_node(parent_node))
        else:
            return parent_node

    def get_child_node(self, _node: py2neo.Node, ignore_error_flag=False) -> py2neo.Node:
        """
        get first child node

        :param ignore_error_flag:
        :param _node:
        :return:
        """
        return self.get_ith_child_node(_node, ignore_error_flag=ignore_error_flag)

    def get_ith_parent_node(self, _node: py2neo.Node, i: int = 0, ignore_error_flag=False) -> py2neo.Node or None:
        """
        res -[:PARENT_OF]-> _node

        查找节点的第i个parent node 节点

        :param ignore_error_flag:
        :param _node:
        :param i:
        :return:
        """
        _node_cp = self.find_parent_nodes(_node)
        if i <= (_node_cp.__len__() - 1):
            return _node_cp[i]
        else:
            if ignore_error_flag:
                return None
            else:
                raise Neo4jNodeListIndexError(buffer=_node_cp, index=i)

    def get_control_node_condition(self, _node: py2neo.Node, ignore_error=False) -> py2neo.Node:
        """
        get the control condition of control node

        :param _node:
        :return:
        """
        if not ignore_error:
            assert _node[NODE_TYPE] in {TYPE_IF, TYPE_IF_ELEM, TYPE_WHILE, TYPE_DO_WHILE}
        else:
            if _node[NODE_TYPE] not in {TYPE_IF, TYPE_IF_ELEM, TYPE_WHILE, TYPE_DO_WHILE}:
                return _node
                # , TYPE_FOR, TYPE_FOREACH NOT CONSIDER
        if _node[NODE_TYPE] in {TYPE_WHILE, TYPE_DO_WHILE, TYPE_IF_ELEM}:
            return self.get_ith_child_node(_node, 0)
        if _node[NODE_TYPE] in {TYPE_IF}:
            return self.get_ith_child_node(
                self.get_ith_child_node(_node, 0), 0
            )

    def get_ith_child_node(self, _node: py2neo.Node, i: int = 0, ignore_error_flag=False) -> py2neo.Node or None:
        """
        _node -[:PARENT_OF]-> res

        查找节点的第i个 child node 节点

        :param _node:
        :param i:
        :return:
        """
        _node_cp = self.find_child_nodes(_node)
        if i <= (_node_cp.__len__() - 1):
            return _node_cp[i]
        else:
            if ignore_error_flag:
                return None
            else:
                raise Neo4jNodeListIndexError(buffer=_node_cp, index=i)

    def get_node_itself(self, _id: int):
        """
        in higher neo4j , the batch import will make id not match the result identity ,
        here I use the slower function instead

        :param _id:
        :return:
        """
        if self.syn_flag is None:
            slow = self.__get_node_itself_slow(_id)
            fast = self.__get_node_itself_fast(_id)
            if fast is not None and slow.identity == fast.identity:
                self.syn_flag = True
            else:
                self.syn_flag = False
            # logger.info(f"Syn static compiled , collect with syn_flag={self.syn_flag} ")

        # load cache ! more faster !
        if _id in self.__node_itself_cache.keys():
            return self.__node_itself_cache[_id]
        else:
            node = self.__get_node_itself_fast(_id) if self.syn_flag else self.__get_node_itself_slow(_id)
            self.__node_itself_cache[_id] = node
            return node

    def get_node_itself_by_identity(self, _id: int):
        return self.__get_node_itself_fast(_id)

    def __get_node_itself_fast(self, _id: int):
        return self.neo4j_graph.nodes.get(identity=_id)

    def __get_node_itself_slow(self, _id: int):
        return self.neo4j_graph.nodes.match(id=_id).limit(1).first()

    def _get_node_belong_file_node(self, _node: py2neo.Node) -> py2neo.Node:
        """
        返回_node所属的文件node OLD METHOD

        :param _node:
        :return:
        """
        tmp = _node
        while tmp:
            _node = tmp
            tmp = self.get_parent_node(_node, ignore_error_flag=True)
        return _node

    def get_node_belong_file_node(self, _node: py2neo.Node) -> py2neo.Node:
        """


        :param _node:
        :return: return the File name node belongs to this node
        """
        if NODE_FILEID in _node.keys():
            file_system_node = self.match(LABEL_FILESYSTEM, id=_node[NODE_FILEID]).first()
            return self.get_node_from_file_system(file_system_node)
        else:
            return self._get_node_belong_file_node(_node)

    def get_node_belong_file_name(self, _node: py2neo.Node) -> str:
        """
        返回_node所属的文件名

        :param _node:
        :return:
        """
        return self.get_node_belong_file_node(_node)[NODE_NAME]

    def get_file_name_belong_node(self, _file_name: str, match_strategy=1) -> Union[py2neo.Node, None]:
        """
        return file name belong node name

        :param _file_name:
        :param match_strategy:
            1 (default)     use `in` case to match
            0               use `=` case to match
        :return:
        """
        if match_strategy == 1:
            nodes = [i for i in
                     self.match("AST", ).where(f"_.type='{TYPE_TOPLEVEL}' and   _.name CONTAINS '{_file_name}' ")]
            if nodes.__len__() >= 1:  # FOR limesurvey the version and real_version name is not complete .... we need to refine it
                best_index = StringMatcher.match_best_similar_str_index(_file_name, [i[NODE_NAME] for i in nodes])
                return nodes[best_index]
            else:
                return None  # file not found error;
        elif match_strategy == 0:
            return self.match(LABEL_AST, ).where(f"_.type='{TYPE_TOPLEVEL}' and '{_file_name}' =  _.name").limit(
                1).first()

    def get_node_from_file_system(self, _node: py2neo.Node) -> py2neo.Node:
        """
        given

        :param _node:
        :return:
        """
        r = self.neo4j_graph.relationships.match(nodes=[_node], r_type=FILE_EDGE).first()
        return r.end_node

    def get_toplevel_file_node_from_file_name(self, file_name,
                                              verbose=False,
                                              compare_method=Levenshtein.ratio) -> py2neo.Node:
        """
        input file name
        :param file_name:
        :return:
        """
        base_file_name = os.path.basename(file_name)
        selector = []
        selector_string = []
        for file_system_node in self.match(LABEL_FILESYSTEM, name=base_file_name):
            res_node = self.get_node_from_file_system(file_system_node)
            selector.append(res_node.identity)
            selector_string.append(res_node[NODE_NAME])
            if verbose:
                logger.info(res_node)
        if selector_string.__len__() >= 1 and selector_string[0].split("/")[0].__len__() >= 40:
            selector_string = [
                i[i.index("/"):] for i in selector_string
            ]
        else:
            logger.warning("Empty file for " + file_name)
            return None
        index = StringMatcher.match_best_similar_str_index(file_name, selector_string, method=compare_method)
        return self.get_node_itself(selector[index])

    def build_cfg_subgraph(self, node: py2neo.Node, upper_node_id: int = 0xffff) -> nx.DiGraph:
        """
        NOW our cfg graph is well-armed !

        :param node:
        :return:
        """
        cfg_subgraph = nx.DiGraph()
        cfg_subgraph.add_node(node[NODE_INDEX], lineno=node[NODE_LINENO], node_id=node[NODE_INDEX],
                              node_type=node[NODE_TYPE])
        assert self.check_apoc_active()
        # TODO This language should be impprove
        relationships_list = [i for i, in self.query(f"""
        MATCH (p:AST {{ id:{node[NODE_INDEX]}  }} )
        CALL apoc.path.subgraphAll(p, {{
        relationshipFilter: 'FLOWS_TO>',
        labelFilter: '+AST',
        minLevel: 0
        }})
        YIELD nodes, relationships
        RETURN [rel in relationships WHERE type(rel) = 'FLOWS_TO']
        """)][-1]
        for relationships in relationships_list:  # type:py2neo.Relationship
            if relationships.end_node[NODE_INDEX] > upper_node_id:
                continue
            __node = relationships.start_node
            cfg_subgraph.add_node(__node[NODE_INDEX],
                                  lineno=__node[NODE_LINENO],
                                  node_id=__node[NODE_INDEX],
                                  node_type=__node[NODE_TYPE])
            __node = relationships.end_node
            cfg_subgraph.add_node(__node[NODE_INDEX],
                                  lineno=__node[NODE_LINENO],
                                  node_id=__node[NODE_INDEX],
                                  node_type=__node[NODE_TYPE])
            cfg_subgraph.add_edge(relationships.start_node[NODE_INDEX], relationships.end_node[NODE_INDEX],
                                  add_rels="CFG", **{i: relationships[i] for i in relationships.keys()})
        return cfg_subgraph

    def build_stmt_subgraph(self, node: py2neo.Node) -> nx.DiGraph:
        """


        :param node:
        :return:
        """
        assert node['type'] in {TYPE_TOPLEVEL, TYPE_METHOD, TYPE_FUNC_DECL}
        input_node = self.search_child_nodes(node, include_type=[TYPE_STMT_LIST])[-1]

        def __add_node_and_edge(p_node, c_node):
            res_graph.add_node(
                c_node.identity, lineno=c_node['lineno'], type=c_node['type'],
                node_id=c_node.identity
            )
            res_graph.add_edge(
                p_node.identity, c_node.identity
            )

        def __join_node_which_has_stmt_list_types(p_node, ):
            for c_node in self.search_child_nodes(p_node, include_type=[TYPE_STMT_LIST]):
                __add_node_and_edge(p_node, c_node)
            return self.search_child_nodes(p_node, include_type=[TYPE_STMT_LIST])

        def __visit(current_node: py2neo.Node):
            assert current_node['type'] == TYPE_STMT_LIST
            rt_list = []
            res_graph.add_node(
                current_node.identity, lineno=current_node['lineno'], type=current_node['type'],
                node_id=current_node.identity
            )
            for __node in self.find_child_nodes(current_node):
                __add_node_and_edge(p_node=current_node, c_node=__node)
                if __node['type'] in HAS_STMT_LIST_TYPES:
                    rt_list.extend(
                        __join_node_which_has_stmt_list_types(p_node=__node)
                    )
                if __node['type'] in CHILD_HAS_STMT_LIST_TYPES:
                    if __node['type'] == "AST_IF":
                        for __node_child in self.find_child_nodes(__node):
                            __add_node_and_edge(p_node=__node, c_node=__node_child)
                            if __node_child['type'] in HAS_STMT_LIST_TYPES:
                                rt_list.extend(
                                    __join_node_which_has_stmt_list_types(p_node=__node_child)
                                )
                    elif __node['type'] == "AST_SWITCH":
                        _switch_list_node = self.find_child_nodes(__node)[-1]
                        # NOT PAY ATTENTION TO SWITCH_LIST NODE !
                        for __node_child in self.find_child_nodes(_switch_list_node):
                            __add_node_and_edge(p_node=__node, c_node=__node_child)
                            if __node_child['type'] in HAS_STMT_LIST_TYPES:
                                rt_list.extend(
                                    __join_node_which_has_stmt_list_types(p_node=__node_child)
                                )
            return rt_list

        res_graph = nx.DiGraph()
        q = Queue()
        q.put(input_node)
        while not q.empty():
            _node = q.get()
            _res = __visit(_node)
            for _r in _res:
                q.put(_r)
        return res_graph

    @SortWrapper()
    def find_all_child_nodes(self, _node: py2neo.Node, max_depth=20, not_include_self: bool = False,
                             node_type_filter: Union[List[str], str, Set[str]] = None) -> List[py2neo.Node]:
        """
        find all the child node via AST EDGE

        :param node_type_filter:
        :param max_depth: the max depth
        :param _node:
        :param not_include_self: is result contain input node it self ?
        :return:
        """
        query = f"MATCH (A:AST{{id:{_node[NODE_INDEX]}}})-[:PARENT_OF*{not_include_self.__int__()}..{max_depth}]->(B:AST) "
        if isinstance(node_type_filter, str):
            node_type_filter = [node_type_filter]
        elif isinstance(node_type_filter, Set):
            node_type_filter = [i for i in node_type_filter]
        if node_type_filter is not None:
            query += f" WHERE B.type in {node_type_filter.__str__()}"
        return [b for b, in self.neo4j_graph.run(
            query + " RETURN B;"
        )]

    def build_ast_subgraph(self, _node: py2neo.Node, max_depth=100, not_include_self: bool = False) -> py2neo.Subgraph:
        """
        find all the child node via AST EDGE

        :param max_depth: the max depth
        :param _node:
        :param not_include_self: is result contain input node it self ?
        :return:
        """
        path = None
        for i, in self.query(
                f"MATCH p = (A:AST)-[r:PARENT_OF*{not_include_self.__int__()}..{max_depth}]->(B:AST)"
                f" WHERE A.id={_node[NODE_INDEX]}  return p;"):  # type: py2neo.Path
            if path is None:
                path = i
            else:
                path |= i
        return path

    @SortWrapper()
    def find_parent_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        查找节点的所有所有 parent nodes

        :param _node:
        :return:
        """
        return [i.start_node for i in self.neo4j_graph.relationships.match(nodes=[None, _node], r_type=AST_EDGE, )]

    @SortWrapper()
    def find_child_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        查找节点的所有所有 child nodes

        :param _node:
        :return:
        """
        return [i.end_node for i in self.neo4j_graph.relationships.match(nodes=[_node, None], r_type=AST_EDGE, )]

    @SortWrapper()
    def find_function_call_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        (_node:AST) -[:CALLS]-> (_result:AST)

        根据CALLS关系查找callee
        :param _node:
        :return:
        """
        return [i.end_node for i in self.neo4j_graph.relationships.match(nodes=[_node, None], r_type=CALLS_EDGE, )]

    def find_caller(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        寻找函数的调用体
        -> return node function a()
        -> input node  a()

        """
        return self.find_function_call_nodes(_node)

    def find_callee(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        寻找函数的调用位置
        -> input node   function a()
        -> return node  a()
        """
        return self.find_function_define_nodes(_node)

    @SortWrapper()
    def find_function_define_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        (_result:AST) -[:CALLS]-> (_node:AST)
        根据CALLS关系反查函数定义节点

        查找节点的callee

        :param _node:fr
        :return:
        """
        return [i.start_node for i in self.neo4j_graph.relationships.match(nodes=[None, _node], r_type=CALLS_EDGE, )]

    @SortWrapper()
    def find_reach_to_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        (_node:AST) -[:REACHES_TO]-> (_result:AST)

        查找节点的所有所有 PDG后继节点

        :param _node:
        :return:
        """
        return [i.end_node for i in self.neo4j_graph.relationships.match(nodes=[_node, None], r_type=DATA_FLOW_EDGE, )]

    @SortWrapper()
    def find_which_reach_to_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        (_result:AST) -[:REACHES_TO]-> (_node:AST)

        查找节点的所有所有 child nodes

        :param _node:
        :return:
        """
        return [i.start_node for i in
                self.neo4j_graph.relationships.match(nodes=[None, _node], r_type=DATA_FLOW_EDGE, )]

    @SortWrapper()
    def find_flows_to_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        (_node:AST) -[:FLOWS_TO]-> (_result:AST)

        查找节点的所有所有 PDG后继节点

        :param _node:
        :return:
        """
        return [i.end_node for i in self.neo4j_graph.relationships.match(nodes=[_node, None], r_type=CFG_EDGE, )]

    @SortWrapper()
    def find_which_flows_to_nodes(self, _node: py2neo.Node) -> List[py2neo.Node]:
        """
        (_result:AST) -[:FLOWS_TO]-> (_node:AST)

        查找节点的所有所有 child nodes

        :param _node:
        :return:
        """
        return [i.start_node for i in self.neo4j_graph.relationships.match(nodes=[None, _node], r_type=CFG_EDGE, )]

    def find_node_structure(self, _node: py2neo.Subgraph, ) -> List[str]:
        pass

    @UnionWrapper()
    def find_all_vars(self, _node: py2neo.Node, target_type: Union[List, Set] = None, using_cache=True) -> List[str]:
        """
        Given node extract all possible args from this node , return string instead

        :param _node:
        :param target_type:
        :return:
        """
        if using_cache \
                and target_type is None \
                and self.__cache.get(self.find_all_vars.__name__ + _node.identity.__str__()) is not None:
            return self.__cache.get(self.find_all_vars.__name__ + _node.identity.__str__())

        if target_type is None:
            target_type = VAR_TYPES
            result = self.find_all_child_nodes(_node=_node, node_type_filter=target_type)
            _res = [_ for _ in map(self.get_general_node_code, result)]
            self.__cache[self.find_all_vars.__name__ + _node.identity.__str__()] = _res
            return _res
        else:
            result = self.find_all_child_nodes(_node=_node, node_type_filter=target_type)
            return [_ for _ in map(self.get_general_node_code, result)]

    def _find_parent_relationships(self, _node) -> List[py2neo.Relationship]:
        """
        查找与节点存在 PARENT_OF 关系的节点，返回RELATIONSHIP 列表

        :param _node:
        :return:
        """
        return list(self.neo4j_graph.relationships.match(nodes={_node}, r_type=AST_EDGE))

    def _find_flows_to_relationships(self, _node) -> List[py2neo.Relationship]:
        """
        查找与节点存在 PARENT_OF 关系的节点，返回RELATIONSHIP 列表

        :param _node:
        :return:
        """
        return list(self.neo4j_graph.relationships.match(nodes={_node}, r_type=CFG_EDGE))

    def _find_reaches_relationships(self, _node) -> List[py2neo.Relationship]:
        """
        查找与节点存在 REACHES 关系的节点，返回RELATIONSHIP 列表

        :param _node:
        :return:
        """
        return list(self.neo4j_graph.relationships.match(nodes={_node}, r_type=DATA_FLOW_EDGE))

    def _find_calls_relationships(self, _node) -> List[py2neo.Relationship]:
        """
        查找与节点存在 CALLS 关系的节点，返回RELATIONSHIP 列表

        :param _node:
        :return:
        """
        return list(self.neo4j_graph.relationships.match(nodes={_node}, r_type=CALLS_EDGE))

    def get_toplevel_file_first_statement(self, toplevel_file_node):
        assert toplevel_file_node[NODE_TYPE] in {TYPE_TOPLEVEL} and \
               NODE_FLAGS in toplevel_file_node.keys() and \
               set(toplevel_file_node[NODE_FLAGS]) & {FLAG_TOPLEVEL_FILE}
        stmt = self.get_child_node(toplevel_file_node)
        return self.get_child_node(stmt)

    def get_function_declare_node_first_statement(self, function_declare_node) -> py2neo.Node:
        assert function_declare_node[NODE_TYPE] in FUNCTION_DECLARE_TYPES
        stmt = self.get_ith_child_node(function_declare_node, 2)
        first_node_ast = self.get_child_node(stmt)
        return first_node_ast

    def get_function_declare_node_exit_statements(self, function_declare_node) -> List[py2neo.Node]:
        # adhoc
        assert function_declare_node[NODE_TYPE] in FUNCTION_DECLARE_TYPES
        stmt = self.get_ith_child_node(function_declare_node, 2)
        for cut in range(0, 10):
            last_node_ast = self.match_first(lineno=function_declare_node[NODE_ENDLINENO] - cut,
                                             fileid=function_declare_node[NODE_FILEID])
            if last_node_ast is not None:
                break
        return [last_node_ast]

    def get_function_arg_ith_node(self, node: py2neo.Node, i=0) -> py2neo.Node:
        """
        get the ith arg of the node

        :param node:
        :param i:
        :return:
        """
        if node[NODE_TYPE] in {TYPE_EXIT, TYPE_ECHO,
                               TYPE_INCLUDE_OR_EVAL,
                               TYPE_PRINT, TYPE_RETURN,
                               TYPE_UNSET, TYPE_ISSET}:  # 特殊arg
            arg_list = self.find_child_nodes(node)
            return arg_list[i]
        arg_list = self.find_function_arg_node_list(node)
        if arg_list.__len__() == 0:
            logger.fatal(f"warning {node} don't have ARG LIST")
        else:
            try:
                return arg_list[i]
            except IndexError as e:
                logger.warning(f"got {e}  system will return None instead")
                return None

    def find_function_arg_node_list(self, node: py2neo.Node) -> List[py2neo.Node]:
        """
        find the arg list from the node, so the input node type must be...

        :param node:
        :return:
        """
        if node[NODE_TYPE] in {TYPE_INCLUDE_OR_EVAL, TYPE_ECHO, TYPE_PRINT, TYPE_EXIT, TYPE_METHOD, TYPE_RETURN}:
            return self.find_child_nodes(node)
        return self.find_child_nodes(
            self.search_child_nodes(node, include_type=[TYPE_ARG_LIST])[0]
        )

    def get_function_arg_node_cnt(self, node: py2neo.Node) -> int:
        """
        get the arg list length , and return its length  ; default 1

        :param node:
        :return:
        """
        if node[NODE_TYPE] in {TYPE_EXIT, TYPE_ECHO, TYPE_INCLUDE_OR_EVAL, TYPE_PRINT, TYPE_RETURN}:  # 特殊arg
            return 1
        arg_list = self.find_function_arg_node_list(node)
        if arg_list.__len__() == 0:
            logger.fatal(f"warning {node.__str__()} don't have ARG LIST")
            return 0  # return 0 ?
        else:
            try:
                return arg_list.__len__()
            except IndexError as e:
                logger.warning(f"got {e} for {node.__str__()}  system will return None instead")
                return 1

    def get_general_node_code(self, node: py2neo.Node, using_cache=True, ) -> str:
        """
        a very awesome method to convert node code quickly without TAC engine

        :param node:
        :param using_cache:
        :return:
        """
        if using_cache \
                and self.__cache.get(self.get_general_node_code.__name__ + node.identity.__str__()) is not None:
            return self.__cache.get(self.get_general_node_code.__name__ + node.identity.__str__())

        code = None
        if node[NODE_TYPE] == TYPE_EXIT:
            code = "die"
        elif node[NODE_TYPE] == TYPE_FUNC_DECL:
            code = node[NODE_NAME]
        elif node[NODE_TYPE] == TYPE_ISSET:
            code = "isset"
        elif node[NODE_TYPE] == TYPE_ECHO:
            code = "echo"
        elif node[NODE_TYPE] == TYPE_PRINT:
            code = "print"
        elif node[NODE_TYPE] == TYPE_RETURN:
            code = "return"
        elif node[NODE_TYPE] == TYPE_UNSET:
            code = "unset"
        elif node[NODE_TYPE] == TYPE_EMPTY:
            code = "empty"
        elif node[NODE_TYPE] == TYPE_INCLUDE_OR_EVAL and (
                set(node[NODE_FLAGS]) & {FLAG_EXEC_INCLUDE, FLAG_EXEC_INCLUDE_ONCE, FLAG_EXEC_REQUIRE,
                                         FLAG_EXEC_REQUIRE_ONCE}):
            if set(node[NODE_FLAGS]) & {FLAG_EXEC_INCLUDE}:
                code = "include"
            if set(node[NODE_FLAGS]) & {FLAG_EXEC_INCLUDE_ONCE}:
                code = "include_once"
            if set(node[NODE_FLAGS]) & {FLAG_EXEC_REQUIRE}:
                code = "require"
            if set(node[NODE_FLAGS]) & {FLAG_EXEC_REQUIRE_ONCE}:
                code = "require_once"
        elif node[NODE_TYPE] == TYPE_BREAK:
            code = "break"
        elif node[NODE_TYPE] == TYPE_INCLUDE_OR_EVAL and (set(node[NODE_FLAGS]) & {FLAG_EXEC_EVAL}):
            code = "eval"
        elif node[NODE_TYPE] == TYPE_NULL:
            code = 'null'

        if code is not None:
            self.__cache[self.get_general_node_code.__name__ + node.identity.__str__()] = code
            return code

        if "get_{}_code".format(node[NODE_TYPE].lower()) not in self.__class_method:
            if node[NODE_TYPE] not in {
                TYPE_ASSIGN, TYPE_ASSIGN_OP, TYPE_POST_INC, TYPE_ASSIGN_REF, TYPE_PRE_INC
            }:
                logger.warning(f"call no reg function with node type {node[NODE_TYPE]}")
            code = f"NOT_SUPPORT_FOR_{node[NODE_TYPE]}"
        else:
            code = eval("self.get_{}_code(node)".format(node[NODE_TYPE].lower()))
        self.__cache[self.get_general_node_code.__name__ + node.identity.__str__()] = code
        return code

    def get_ast_method_code(self, node: py2neo.Node) -> str:
        """
        AST_METHOD
        """
        assert node[NODE_TYPE] == TYPE_METHOD
        return node[NODE_NAME]

    def get_ast_function_decl_code(self, node: py2neo.Node) -> str:
        """
        AST_FUNC_DECL
        """
        assert node[NODE_TYPE] == TYPE_FUNC_DECL
        return node[NODE_NAME]

    def get_ast_new_code(self, node: py2neo.Node) -> str:
        assert node[NODE_TYPE] == TYPE_NEW  # 构造函数
        if self.get_child_node(self.get_child_node(node))[NODE_CODE] is not None:
            return self.get_child_node(self.get_child_node(node))[NODE_CODE]
        elif self.get_child_node(node)[NODE_TYPE] == TYPE_PROP:
            return self.get_ast_prop_code(self.get_child_node(node))
        else:
            raise Neo4jQuickCodeGenerationError(TYPE_NEW + node.__str__())

    def get_ast_var_code(self, node: py2neo.Node) -> str:
        assert node[NODE_TYPE] == TYPE_VAR
        # k = [ i for i in self.get_child_node(node).keys()]
        # k = [ i for i in self.get_child_node(node).keys()]
        # if NODE_CODE not in k:
        #     return '$' + "{$" + self.get_child_node(self.get_child_node(node))[NODE_CODE] + "}"
        if NODE_CODE in self.get_child_node(node).keys():
            return '$' + self.get_child_node(node)[NODE_CODE]
        elif NODE_CODE in self.get_child_node(self.get_child_node(node)).keys():
            # AST_VAR->AST_VAR->AST_STRING REPRESENTS FOR FORM LIKE $$A
            # Only support for $a and $$a ； for $$$a we will not support it
            return '$$' + self.get_child_node(self.get_child_node(node))[NODE_CODE]
        else:
            return '$uk'

    def __node_type_assertion(self, node: py2neo.Node, TYPE):
        assert node[NODE_TYPE] == TYPE
        return True

    def get_ast_const_code(self, node: py2neo.Node) -> str:
        assert node[NODE_TYPE] == TYPE_CONST
        return self.get_child_node(self.get_child_node(node))[NODE_CODE]

    def get_ast_prop_code(self, node: py2neo.Node) -> str:
        """
        get the function name of the ast pro node ,so the input node type must be AST_PROP

        :param node:
        :return:
        """
        assert node[NODE_TYPE] == TYPE_PROP
        attribute = self.__normalize(self.get_ith_child_node(node, -1)[NODE_CODE])
        clazz = self.get_child_node(self.get_ith_child_node(node, 0))[NODE_CODE]
        return f"${clazz}->{attribute}"

    def get_ast_static_prop_code(self, node: py2neo.Node) -> str:
        """
        get the function name of the ast pro node ,so the input node type must be AST_PROP

        :param node:
        :return:
        """
        assert node[NODE_TYPE] == TYPE_STATIC_PROP
        attribute = self.__normalize(self.get_ith_child_node(node, -1)[NODE_CODE])
        clazz = self.get_child_node(self.get_ith_child_node(node, 0))[NODE_CODE]
        return f"{clazz}::${attribute}"

    def get_ast_dim_body_code(self, node: py2neo.Node) -> str:
        """
        同样，这个函数目标只是提出DIM的第一层结果。并且不包含$，主要用来检测全局变量
        """
        assert node[NODE_TYPE] == TYPE_DIM
        dim_body = self.get_child_node(self.get_child_node(node))[NODE_CODE]
        return dim_body

    # def get_ast_dim_structure_code(self, node: py2neo.Node) -> Tuple[str, List[str]]:
    #     """
    #     返回 数组的各个部分，弥补get_ast_dim_code的缺陷，主要方法是把它的每个部分拆开来，分别喂到 general code中，并根据返回结果进行解析
    #     """
    #     assert node[NODE_TYPE] == TYPE_DIM
    #     dim_body = self.get_child_node(self.get_child_node(node))[NODE_CODE]
    #
    #     return dim_body

    def get_ast_dim_code(self, node: py2neo.Node) -> str:
        """
        TODO BIG PROBLEM
        get the code of the ast dim node ,so the input node type must be AST_DIM
        这个函数不支持循环数组，比如说 $SHAOBAO['ASDF']['SDFA']['FDSA]这样的。并且数组的SLICE不能太难，总归能解决一大部分问题了。

        :param node:
        :return:
        """
        assert node[NODE_TYPE] == TYPE_DIM
        dim_body = self.get_general_node_code(self.get_ith_child_node(node, 0))
        dim_slice = self.get_general_node_code(self.get_ith_child_node(node, 1))
        if dim_slice == 'null':
            dim_slice = ''
        elif self.get_ith_child_node(node, 1)[NODE_TYPE] == TYPE_STRING:
            dim_slice = f"\"{dim_slice}\""
        logger.info(f"DEBUGGING AST_DIM, {self.get_ith_child_node(node, 1)[NODE_TYPE]} => {dim_body}[{dim_slice}] ")
        return f"{dim_body}[{dim_slice}]"

    def get_ast_call_code(self, node: py2neo.Node) -> str:
        """
        get the function name of the ast call node ,so the input node type must be AST_CALL

        :param node:
        :return:
        """
        assert node[NODE_TYPE] == TYPE_CALL
        return self.get_child_node(self.get_child_node(node))[NODE_CODE]

    def get_ast_static_call_code(self, node: py2neo.Node) -> str:
        """
        get the function name of the ast static call node , so the input node type must be AST_STATIC_CALL

        :param node:
        :return:
        """
        assert node[NODE_TYPE] == TYPE_STATIC_CALL
        ch_nodes = self.find_child_nodes(node)
        class_name_expr = self.get_child_node(ch_nodes[0])[NODE_CODE]
        if ch_nodes.__len__() >= 1:
            class_method_expr = ch_nodes[1][NODE_CODE]
        else:
            # like shaobao::$a() 动态的方法
            class_method_expr = "$" + self.get_child_node(self.get_ith_child_node(node, 1))[NODE_CODE]
        return f"{class_name_expr}::{class_method_expr}"

    def get_ast_class_const_code(self, node: py2neo.Node) -> str:
        assert node[NODE_TYPE] == TYPE_CLASS_CONST
        ch_nodes = self.find_child_nodes(node)
        class_name_expr = self.get_child_node(ch_nodes[0])[NODE_CODE]
        if ch_nodes.__len__() >= 1:
            class_method_expr = ch_nodes[1][NODE_CODE]
        else:
            # like shaobao::$a 方法
            class_method_expr = "$" + self.get_child_node(self.get_ith_child_node(node, 1))[NODE_CODE]
        return f"{class_name_expr}::{class_method_expr}"

    def get_ast_method_call_code(self, node: py2neo.Node) -> str:
        """
        get the function name of the ast method call node , so the input node type must be AST_METHOD_CALL

        :param node:
        :return:
        """
        assert node[NODE_TYPE] == TYPE_METHOD_CALL
        if self.get_ith_child_node(node, 1)[NODE_TYPE] == TYPE_VAR:
            return self.get_ith_child_node(self.get_ith_child_node(node, 1), 0)[NODE_CODE]
        return self.get_ith_child_node(node, 1)[NODE_CODE]

    def get_function_defined_node_by_name(self, name: str):
        """
        大概会执行如下语句：

        MATCH (A:AST) WHERE A.type in ["AST_METHOD","AST_FUNC_DECL"] AND A.name = "get_post" return A;

        :return: Node List
        """
        if "new " in name:
            name = name.replace("new", "").strip()
            return self.get_class_construct_function(
                self.get_class_defined_node_by_name(name)
            )
        return self.neo4j_graph.nodes.match("AST").where(
            f"_.name='{name}' and _.type in ['AST_METHOD','AST_FUNC_DECL']"
        ).first()

    def get_class_defined_node_by_name(self, name: str):
        """
        根据类名，找到类定义节点

        :param name:
        """
        return self.neo4j_graph.nodes.match("AST").where(
            f"_.name='{name}' AND _.type='AST_CLASS'"
        ).first()

    def get_class_construct_function(self, node: py2neo.Node):
        """
        根据类定义节点，找到类构造方法

        :param node:
        """
        class_top_level_node = self.search_child_nodes(node, include_type=[TYPE_TOPLEVEL])[0]
        class_stmt_list_node = self.search_child_nodes(class_top_level_node, include_type=[TYPE_STMT_LIST])[0]
        for i in self.search_child_nodes(class_stmt_list_node, include_type=[TYPE_METHOD]):
            if i[NODE_NAME] == "__construct":
                return i
        return None

    # def find_nodes_belongs_to_funcid(self, funcid: int, SPECIFIED_TYPE=None):
    #     if SPECIFIED_TYPE is None:
    #         SPECIFIED_TYPE = []
    #     return {node.identity: node
    #             for node in self.neo4j_graph.nodes.match("AST").where(
    #             f"_.funcid={funcid} and _.type in {SPECIFIED_TYPE}"
    #         )}

    def is_loop_instance_node(self, _node: py2neo.Node) -> Union[bool, str]:
        """
        judge whether the root node is LOOP instance node
        :return: return LOOP_INSTANCE_TYPE including AST_FOREACH,AST_FOR,AST_WHILE,AST_DO_WHILE
                 or return false
        """
        # note: not implement DO WHILE
        if isinstance(_node, int):
            _node = self.get_node_itself(_node)
        if _node[NODE_TYPE] == TYPE_FOREACH:
            LOOP_TYPE = TYPE_FOREACH
        elif self.get_parent_node(_node)[NODE_TYPE] in {TYPE_FOR} and _node[NODE_CHILDNUM] == 1:
            LOOP_TYPE = TYPE_FOR
        elif self.get_parent_node(_node)[NODE_TYPE] in {TYPE_WHILE}:
            LOOP_TYPE = TYPE_WHILE
        else:
            LOOP_TYPE = False
        return LOOP_TYPE

    def has_flows_to_import(self, _node: py2neo.Node) -> bool:
        """
        judge if the input node has CFG to others node

        :param _node:
        :return:
        """
        return self.find_which_flows_to_nodes(_node).__len__().__bool__()

    def has_cfg_export(self, _node: py2neo.Node) -> bool:
        """
        judge if other nodes has CFG to the input node

        :param _node:
        :return:
        """
        return self.find_flows_to_nodes(_node).__len__().__bool__()

    def has_cfg(self, _node: py2neo.Node) -> bool:
        """
        judge if the input node has CFG

        :param _node:
        :return:
        """
        return [i.start_node for i in
                self.neo4j_graph.relationships.match(nodes={_node}, r_type=CFG_EDGE, )].__len__().__bool__()

    # noinspection PyTypeChecker
    def get_general_node_range(self, node: py2neo.Node, use_cache=True) -> Tuple[int, int]:
        parent_node = self.get_parent_node(node)
        node_hash = f"{self.get_general_node_range.__name__}{node.identity.__str__()} "
        if node_hash not in self.__cache.keys():
            if node[NODE_TYPE] in {TYPE_FOREACH}:
                end_id = self.find_flows_to_nodes(node)[-1][NODE_INDEX]
                if end_id > node[NODE_INDEX]:
                    node_range = (node[NODE_INDEX], end_id)
                    self.__cache[node_hash] = node_range
                else:
                    raise NotImplementedError()
            elif node[NODE_TYPE] in {TYPE_WHILE}:
                pass
                raise NotImplementedError()
            elif parent_node[NODE_TYPE] in {TYPE_IF_ELEM}:
                # example
                # 1: if(a){
                # 2:   print(A)}
                # 3: elif(b){
                # 4:   print(B)}
                # 5: elif(c){
                # 6:   print(C)}
                # 7: else(d){
                # 8:   print(D)}
                # if the input node is 1:a -> return (1:a,2:A)
                # if the input node is 2:b -> return (2:b,2:B)
                # it is hard for given input 2:b -> return (2:b,8:D)
                logger.info(f"detect the ast root node of {TYPE_IF_ELEM} , use new rule instead {node[NODE_TYPE]}")
                if self.find_flows_to_nodes(node).__len__() == 0:
                    logger.fatal(f"no flows to node for node {node}")
                    node_range = (parent_node[NODE_INDEX], parent_node[NODE_INDEX] + 500)
                else:
                    node_range = (parent_node[NODE_INDEX], self.find_flows_to_nodes(node)[-1][NODE_INDEX] - 1)
                self.__cache[node_hash] = node_range
            else:
                reg_function = [i for i in self.__dir__() if not i.__str__().startswith("_")]
                if "get_{}_node_range".format(node[NODE_TYPE].lower()) not in reg_function:
                    node_range = (node[NODE_INDEX], self.find_all_child_nodes(node)[-1][NODE_INDEX])
                else:
                    node_range = eval("self.get_{}_node_range(node)".format(node[NODE_TYPE].lower()))
                self.__cache[node_hash] = node_range
        return self.__cache[node_hash]

    def get_if_code_range(self, node: py2neo.Node) -> Tuple[int, int]:
        pass

    def get_ast_func_decl_range(self, node: py2neo.Node) -> Tuple[int, int]:
        assert node[NODE_TYPE] == TYPE_FUNC_DECL
        RR = self.match_first(LABEL_AST, **{NODE_FILEID: node[NODE_FILEID], NODE_LINENO: node[NODE_ENDLINENO]})
        if not RR:
            RR = self.match_first(LABEL_AST, **{NODE_FILEID: node[NODE_FILEID], NODE_LINENO: node[NODE_ENDLINENO] - 1})
        if not RR:
            logger.fatal("endlineno is wriong or need enhancement")
        end_line_root_node = self.get_ast_root_node(RR)
        last_node = self.find_all_child_nodes(end_line_root_node)[-1]
        return node[NODE_INDEX], last_node[NODE_INDEX]

    def get_ast_method_range(self, node: py2neo.Node) -> Tuple[int, int]:
        assert node[NODE_TYPE] == TYPE_METHOD
        RR = self.match_first(LABEL_AST, **{NODE_FILEID: node[NODE_FILEID], NODE_LINENO: node[NODE_ENDLINENO]})
        if not RR:
            RR = self.match_first(LABEL_AST, **{NODE_FILEID: node[NODE_FILEID], NODE_LINENO: node[NODE_ENDLINENO] - 1})
        if not RR:
            logger.fatal("endlineno is wriong or need enhancement")
        end_line_root_node = self.get_ast_root_node(RR)
        last_node = self.find_all_child_nodes(end_line_root_node)[-1]
        return node[NODE_INDEX], last_node[NODE_INDEX]
